from qtpy.QtGui import QColor, QColorConstants, QPalette

ColorRole = QPalette.ColorRole
ColorGroup = QPalette.ColorGroup
Disabled = QPalette.ColorGroup.Disabled

Red = QColorConstants.Red
Black = QColorConstants.Black
White = QColorConstants.White
Blue = QColorConstants.Blue

dark = QPalette()
dark.setColor(ColorGroup.Active, ColorRole.WindowText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Active, ColorRole.Button, QColor(0x323232))
dark.setColor(ColorGroup.Active, ColorRole.Light, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Active, ColorRole.Midlight, QColor(0x2A2A2A))
dark.setColor(ColorGroup.Active, ColorRole.Dark, QColor(0x212121))
dark.setColor(ColorGroup.Active, ColorRole.Mid, QColor(0x262626))
dark.setColor(ColorGroup.Active, ColorRole.Text, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Active, ColorRole.BrightText, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Active, ColorRole.ButtonText, QColor(0xAAD4FF))
dark.setColor(ColorGroup.Active, ColorRole.Base, QColor(0x242424))
dark.setColor(ColorGroup.Active, ColorRole.Window, QColor(0x323232))
dark.setColor(ColorGroup.Active, ColorRole.Shadow, QColor(0x191919))
dark.setColor(ColorGroup.Active, ColorRole.Highlight, QColor(0x308CC6))
dark.setColor(ColorGroup.Active, ColorRole.HighlightedText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Active, ColorRole.Link, QColor(0x308CC6))
dark.setColor(ColorGroup.Active, ColorRole.LinkVisited, QColor(0xFF00FF))
dark.setColor(ColorGroup.Active, ColorRole.AlternateBase, QColor(0x2B2B2B))
dark.setColor(ColorGroup.Active, ColorRole.ToolTipBase, QColor(0xFFFFDC))
dark.setColor(ColorGroup.Active, ColorRole.ToolTipText, Black)
dark.setColor(ColorGroup.Active, ColorRole.PlaceholderText, QColor(0xA0A0A0))
dark.setColor(ColorGroup.Active, ColorRole.NoRole, Black)
dark.setColor(ColorGroup.Disabled, ColorRole.WindowText, QColor(0x828282))
dark.setColor(ColorGroup.Disabled, ColorRole.Button, QColor(0x323232))
dark.setColor(ColorGroup.Disabled, ColorRole.Light, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Disabled, ColorRole.Midlight, QColor(0x2A2A2A))
dark.setColor(ColorGroup.Disabled, ColorRole.Dark, QColor(0xBEBEBE))
dark.setColor(ColorGroup.Disabled, ColorRole.Mid, QColor(0x262626))
dark.setColor(ColorGroup.Disabled, ColorRole.Text, QColor(0x828282))
dark.setColor(ColorGroup.Disabled, ColorRole.BrightText, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Disabled, ColorRole.ButtonText, QColor(0x828282))
dark.setColor(ColorGroup.Disabled, ColorRole.Base, QColor(0x323232))
dark.setColor(ColorGroup.Disabled, ColorRole.Window, QColor(0x323232))
dark.setColor(ColorGroup.Disabled, ColorRole.Shadow, QColor(0x252525))
dark.setColor(ColorGroup.Disabled, ColorRole.Highlight, QColor(0x919191))
dark.setColor(ColorGroup.Disabled, ColorRole.HighlightedText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Disabled, ColorRole.Link, QColor(0x308CC6))
dark.setColor(ColorGroup.Disabled, ColorRole.LinkVisited, QColor(0xFF00FF))
dark.setColor(ColorGroup.Disabled, ColorRole.AlternateBase, QColor(0x2B2B2B))
dark.setColor(ColorGroup.Disabled, ColorRole.ToolTipBase, QColor(0xFFFFDC))
dark.setColor(ColorGroup.Disabled, ColorRole.ToolTipText, Black)
dark.setColor(ColorGroup.Disabled, ColorRole.PlaceholderText, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Disabled, ColorRole.NoRole, Black)
dark.setColor(ColorGroup.Inactive, ColorRole.WindowText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Inactive, ColorRole.Button, QColor(0x323232))
dark.setColor(ColorGroup.Inactive, ColorRole.Light, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Inactive, ColorRole.Midlight, QColor(0x2A2A2A))
dark.setColor(ColorGroup.Inactive, ColorRole.Dark, QColor(0x212121))
dark.setColor(ColorGroup.Inactive, ColorRole.Mid, QColor(0x262626))
dark.setColor(ColorGroup.Inactive, ColorRole.Text, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Inactive, ColorRole.BrightText, QColor(0x4B4B4B))
dark.setColor(ColorGroup.Inactive, ColorRole.ButtonText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Inactive, ColorRole.Base, QColor(0x242424))
dark.setColor(ColorGroup.Inactive, ColorRole.Window, QColor(0x323232))
dark.setColor(ColorGroup.Inactive, ColorRole.Shadow, QColor(0x191919))
dark.setColor(ColorGroup.Inactive, ColorRole.Highlight, QColor(0x308CC6))
dark.setColor(ColorGroup.Inactive, ColorRole.HighlightedText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Inactive, ColorRole.Link, QColor(0x308CC6))
dark.setColor(ColorGroup.Inactive, ColorRole.LinkVisited, QColor(0xFF00FF))
dark.setColor(ColorGroup.Inactive, ColorRole.AlternateBase, QColor(0x2B2B2B))
dark.setColor(ColorGroup.Inactive, ColorRole.ToolTipBase, QColor(0xFFFFDC))
dark.setColor(ColorGroup.Inactive, ColorRole.ToolTipText, Black)
dark.setColor(ColorGroup.Inactive, ColorRole.PlaceholderText, QColor(0xF0F0F0))
dark.setColor(ColorGroup.Inactive, ColorRole.NoRole, Black)


light = QPalette()
light.setColor(ColorGroup.Active, ColorRole.WindowText, Black)
light.setColor(ColorGroup.Active, ColorRole.Button, QColor(0xEFEFEF))
light.setColor(ColorGroup.Active, ColorRole.Light, White)
light.setColor(ColorGroup.Active, ColorRole.Midlight, QColor(0xCACACA))
light.setColor(ColorGroup.Active, ColorRole.Dark, QColor(0x9F9F9F))
light.setColor(ColorGroup.Active, ColorRole.Mid, QColor(0xB8B8B8))
light.setColor(ColorGroup.Active, ColorRole.Text, Black)
light.setColor(ColorGroup.Active, ColorRole.BrightText, White)
light.setColor(ColorGroup.Active, ColorRole.ButtonText, Black)
light.setColor(ColorGroup.Active, ColorRole.Base, White)
light.setColor(ColorGroup.Active, ColorRole.Window, QColor(0xEFEFEF))
light.setColor(ColorGroup.Active, ColorRole.Shadow, QColor(0x767676))
light.setColor(ColorGroup.Active, ColorRole.Highlight, QColor(0x308CC6))
light.setColor(ColorGroup.Active, ColorRole.HighlightedText, White)
light.setColor(ColorGroup.Active, ColorRole.Link, Blue)
light.setColor(ColorGroup.Active, ColorRole.LinkVisited, QColor(0xFF00FF))
light.setColor(ColorGroup.Active, ColorRole.AlternateBase, QColor(0xF7F7F7))
light.setColor(ColorGroup.Active, ColorRole.ToolTipBase, QColor(0xFFFFDC))
light.setColor(ColorGroup.Active, ColorRole.ToolTipText, Black)
light.setColor(ColorGroup.Active, ColorRole.PlaceholderText, QColor(0x404040))
light.setColor(ColorGroup.Active, ColorRole.NoRole, Black)
light.setColor(ColorGroup.Disabled, ColorRole.WindowText, QColor(0xBEBEBE))
light.setColor(ColorGroup.Disabled, ColorRole.Button, QColor(0xEFEFEF))
light.setColor(ColorGroup.Disabled, ColorRole.Light, White)
light.setColor(ColorGroup.Disabled, ColorRole.Midlight, QColor(0xCACACA))
light.setColor(ColorGroup.Disabled, ColorRole.Dark, QColor(0xBEBEBE))
light.setColor(ColorGroup.Disabled, ColorRole.Mid, QColor(0xB8B8B8))
light.setColor(ColorGroup.Disabled, ColorRole.Text, QColor(0xBEBEBE))
light.setColor(ColorGroup.Disabled, ColorRole.BrightText, White)
light.setColor(ColorGroup.Disabled, ColorRole.ButtonText, QColor(0xBEBEBE))
light.setColor(ColorGroup.Disabled, ColorRole.Base, QColor(0xEFEFEF))
light.setColor(ColorGroup.Disabled, ColorRole.Window, QColor(0xEFEFEF))
light.setColor(ColorGroup.Disabled, ColorRole.Shadow, QColor(0xB1B1B1))
light.setColor(ColorGroup.Disabled, ColorRole.Highlight, QColor(0x919191))
light.setColor(ColorGroup.Disabled, ColorRole.HighlightedText, White)
light.setColor(ColorGroup.Disabled, ColorRole.Link, Blue)
light.setColor(ColorGroup.Disabled, ColorRole.LinkVisited, QColor(0xFF00FF))
light.setColor(ColorGroup.Disabled, ColorRole.AlternateBase, QColor(0xF7F7F7))
light.setColor(ColorGroup.Disabled, ColorRole.ToolTipBase, QColor(0xFFFFDC))
light.setColor(ColorGroup.Disabled, ColorRole.ToolTipText, Black)
light.setColor(ColorGroup.Disabled, ColorRole.PlaceholderText, QColor(0xC0C0C0))
light.setColor(ColorGroup.Disabled, ColorRole.NoRole, Black)
light.setColor(ColorGroup.Inactive, ColorRole.WindowText, Black)
light.setColor(ColorGroup.Inactive, ColorRole.Button, QColor(0xEFEFEF))
light.setColor(ColorGroup.Inactive, ColorRole.Light, White)
light.setColor(ColorGroup.Inactive, ColorRole.Midlight, QColor(0xCACACA))
light.setColor(ColorGroup.Inactive, ColorRole.Dark, QColor(0x9F9F9F))
light.setColor(ColorGroup.Inactive, ColorRole.Mid, QColor(0xB8B8B8))
light.setColor(ColorGroup.Inactive, ColorRole.Text, Black)
light.setColor(ColorGroup.Inactive, ColorRole.BrightText, White)
light.setColor(ColorGroup.Inactive, ColorRole.ButtonText, Black)
light.setColor(ColorGroup.Inactive, ColorRole.Base, White)
light.setColor(ColorGroup.Inactive, ColorRole.Window, QColor(0xEFEFEF))
light.setColor(ColorGroup.Inactive, ColorRole.Shadow, QColor(0x767676))
light.setColor(ColorGroup.Inactive, ColorRole.Highlight, QColor(0x308CC6))
light.setColor(ColorGroup.Inactive, ColorRole.HighlightedText, White)
light.setColor(ColorGroup.Inactive, ColorRole.Link, Blue)
light.setColor(ColorGroup.Inactive, ColorRole.LinkVisited, QColor(0xFF00FF))
light.setColor(ColorGroup.Inactive, ColorRole.AlternateBase, QColor(0xF7F7F7))
light.setColor(ColorGroup.Inactive, ColorRole.ToolTipBase, QColor(0xFFFFDC))
light.setColor(ColorGroup.Inactive, ColorRole.ToolTipText, Black)
light.setColor(ColorGroup.Inactive, ColorRole.PlaceholderText, Black)
light.setColor(ColorGroup.Inactive, ColorRole.NoRole, Black)
